/* cirparith2.c
 *
 * Does image arithmetic on cirpass MEF fits files
 *
 * Different from cirparith as it's inteded to operate on .ms files,
 * which at the moment don't have dq or time arrays.
 *
 * AJ Dean, 23rd May, 2003
 *
 */

#include <stdlib.h>
#include <string.h>
#include "fitswrap.h" 

/* ITSZERO DIVZERO VARZERO and DIVBYZERO defined in cirpass.h */

#define MOD(X) ( (X)>=0 ? (X) : (-1*(X)) )

int main(int argc, char* argv[])
{

  char* exttypes[]={"SCI","VAR"};
  int   numexts=2;
  int   i,j;
  int   nextend,next;
  int   nextend2;
  long  nx,ny;
  float* operand1[2]; /* [] indexes exts above */
  float* operand2[2];
  float* scidata;
  float* vardata;
  float* data[2];
  float number;

  int imsize=0;
  int naxis=2;
  long naxes[2];

  int divzeroerror=0;

  /* Check for appropriate input */
  if (argc < 5) {
  
      fprintf(stderr, 
          "Usage: %s operand1.fits operator[+|-|x|/] operand2.fits outfile.fits\n", 
           argv[0]);
        exit(1);
  
  } 

  /* find out if second operand is image or number */
  if ( strstr (argv[3], ".fits" ) == NULL &&  
       strstr (argv[3], ".FITS" ) == NULL ){
  
    
    /* No .fits i.e. we were given a number */
    if(sscanf(argv[3],"%f",&number)==0){
      printf("cirparith2: you didn't give me a number or a fits file!\n");
      exit(0);
    }

    /* Read in the data */
    for(i=0; i<numexts; i++){
      
      operand1[i]=read_MS_ME_FITS(argv[1], exttypes[i], &nx, &ny, &nextend);  

    }

    next=nextend/MSNIMSET;
    imsize=nx*ny;
    naxes[0]=nx;
    naxes[1]=ny;
    
    /* Allocate memory */
    if ( (scidata=(float *)malloc(sizeof(float)*imsize*next))  == NULL ||
	 (vardata=(float *)malloc(sizeof(float)*imsize*next))  == NULL ){
      printf("cirparith2: Memory allocation failed\n");
      exit(1);
    }
    
    printf("cirparith2: Read in data, doing the arithmetic\n");
    
    if( strcmp(argv[2],"+")==0 ){
      
      for(j=0; j<imsize*next; j++){
      
	*(scidata+j)=*(operand1[0]+j)+number;
	*(vardata+j)=*(operand1[1]+j);
		
      }
      
    } else if( strcmp(argv[2],"-")==0 ){
      
      for(j=0; j<imsize*next; j++){
	
	*(scidata+j)=*(operand1[0]+j)-number;
	*(vardata+j)=*(operand1[1]+j);
	      
      }
  
    } else if( strcmp(argv[2],"*")==0 ){
      
      for(j=0; j<imsize*next; j++){
	
	*(scidata+j)=*(operand1[0]+j)*number;
	*(vardata+j)=*(operand1[1]+j)*number*number; 
	      
      }
    
    } else if( strcmp(argv[2],"/")==0 ){
    
      for(j=0; j<imsize*next; j++){
      
	if( MOD(number) > ITSZERO ){
	  *(scidata+j)=*(operand1[0]+j) / number;
	  *(vardata+j)=*(operand1[1]+j) / number / number;
	} else {
	  *(scidata+j)=DIVZERO;
	  *(vardata+j)=VARZERO;
	  divzeroerror=1;
	}	
      }
    
    } else {
    
      printf("cirparith2: ERROR! Unrecognised operator!\n");
      exit(1);
    }
    
    data[0]=scidata;
    data[1]=vardata;
        
    /* Write out the new image, using the header and table from the 
       original file */
    printf("cirparith2: Writing out new image\n");
    write_cirp_ms_mef(argv[1], argv[4], exttypes, data, numexts,
		   naxis, naxes, next);
		    
    /* free all dynamically allocated data */
    for(i=0; i<numexts; i++){
      free(operand1[i]);
    }
    
  } else {
    
    /* Read in the data */
    for(i=0; i<numexts; i++){
      
      operand1[i]=read_MS_ME_FITS(argv[1], exttypes[i], &nx, &ny, &nextend);
      
      operand2[i]=read_MS_ME_FITS(argv[3], exttypes[i], &nx, &ny, &nextend2);
   
    }
  
    /* Check the files have the same number of extensions */
    if(nextend!=nextend2){
      printf("cirparith2: ** Warning! Input files do not have the same number of extensions! **\ncirparith2: ** Will just use the first two image extensions. **\n");
      nextend=2;
      nextend2=2;
    }

    next=nextend/MSNIMSET;
    imsize=nx*ny;
    naxes[0]=nx;
    naxes[1]=ny;
    
    /* Allocate memory */
    if ( (scidata=(float *)malloc(sizeof(float)*imsize*next))    == NULL ||
	 (vardata=(float *)malloc(sizeof(float)*imsize*next))  == NULL ){
	
	 printf("cirparith2: Memory allocation failed\n");
      exit(1);
    }
    
    printf("cirparith2: Read in data, doing the arithmetic\n");
    
    if( strcmp(argv[2],"+")==0 ){
      
      for(j=0; j<imsize*next; j++){
      
	*(scidata+j)=*(operand1[0]+j)+*(operand2[0]+j);
	*(vardata+j)=*(operand1[1]+j)+*(operand2[1]+j);
		
      }
      
    } else if( strcmp(argv[2],"-")==0 ){
      
      for(j=0; j<imsize*next; j++){
	
	*(scidata+j)=*(operand1[0]+j)-*(operand2[0]+j);
	*(vardata+j)=*(operand1[1]+j)+*(operand2[1]+j);
	      
      }
  
    } else if( strcmp(argv[2],"*")==0 ){
      
      for(j=0; j<imsize*next; j++){
	
	if( ( MOD( *(operand1[0]+j) ) > ITSZERO ) && 
            ( MOD( *(operand2[0]+j) ) > ITSZERO ) ){
	  *(scidata+j)=*(operand1[0]+j)**(operand2[0]+j);
	  *(vardata+j)=(*(operand1[0]+j) * *(operand2[0]+j)) * 
	             (*(operand1[0]+j) * *(operand2[0]+j)) *
	             (*(operand1[1]+j) / *(operand1[0]+j) / *(operand1[0]+j) + 
		      *(operand2[1]+j) / *(operand2[0]+j) / *(operand2[0]+j));
	 } else {
	   *(scidata+j)=0.0;
	   *(vardata+j)=VARZERO;
	   divzeroerror=1;
	 } 
      
      }
    
    } else if( strcmp(argv[2],"/")==0 ){
    
      for(j=0; j<imsize*next; j++){
      
	if( ( MOD( *(operand1[0]+j) ) > ITSZERO ) && 
            ( MOD( *(operand2[0]+j) ) > ITSZERO ) ){
	  *(scidata+j)=*(operand1[0]+j) / *(operand2[0]+j);
	  *(vardata+j)=(*(operand1[0]+j) / *(operand2[0]+j)) * 
                     (*(operand1[0]+j) / *(operand2[0]+j)) *
	             (*(operand1[1]+j) / *(operand1[0]+j) / *(operand1[0]+j) + 
	              *(operand2[1]+j) / *(operand2[0]+j) / *(operand2[0]+j));
	} else {
	  *(scidata+j)=DIVZERO;
	  *(vardata+j)=VARZERO;
	  divzeroerror=1;
	}	
      
      }
    
    } else {
    
      printf("cirparith2: ERROR! Unrecognised operator!\n");
      exit(1);
    }
    
    data[0]=scidata;
    data[1]=vardata;
       
    /* Write out the new image, using the header and table from the 
       original file */
    printf("cirparith2: Writing out new image\n");
    write_cirp_ms_mef(argv[1], argv[4], exttypes, data, numexts,
		   naxis, naxes, next);
		    
    /* free all dynamically allocated data */
    for(i=0; i<numexts; i++){
      free(operand1[i]);
      free(operand2[i]);
    }
		    	    
  }	  
  
  if(divzeroerror){
      printf("cirparith2: ** Warning! Division by zero occured! **\n");
      printf("cirparith2: ** Output data set to %2.2f, var to %2.2f ",
	     DIVZERO,VARZERO);      
      printf("for affected pixels **\n");
  }
  
  free(scidata);
  free(vardata);
  
  return 0;

}
